/*
 * This file is part of Cockpit.
 *
 * Copyright (C) 2014 Red Hat, Inc.
 *
 * Cockpit is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * Cockpit is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Cockpit; If not, see <http://www.gnu.org/licenses/>.
 */

define([
    "jquery",
    "base1/cockpit"
], function($, cockpit) {
    "use strict";

    var kubernetes = { };

    function debug() {
        if (window.debugging == "all" || window.debugging == "kubernetes")
            console.debug.apply(console, arguments);
    }

    function failure(ex) {
        console.warn(ex);
    }

    function hash(str) {
        var h, i, chr, len;
        if (str.length === 0)
            return 0;
        for (h = 0, i = 0, len = str.length; i < len; i++) {
            chr = str.charCodeAt(i);
            h = ((h << 5) - h) + chr;
            h |= 0; // Convert to 32bit integer
        }
        return Math.abs(h);
    }

    /**
     * HashIndex
     * @size: the number of slots for hashing into
     *
     * A probablisting hash index, where items are added with
     * various keys, and probable matches are returned. Similar
     * to bloom filters, false positives are possible, but never
     * false negatives.
     */
    function HashIndex(size) {
        var self = this;
        var array = [];

        self.add = function add(keys, value) {
            var i, j, p, length = keys.length;
            for (j = 0; j < length; j++) {
                i = hash("" + keys[j]) % size;
                p = array[i];
                if (p === undefined)
                    p = array[i] = [];
                p.push(value);
            }
        };

        self.select = function select(keys) {
            var i, j, interim = [], length = keys.length;
            for (j = 0; j < length; j++) {
                i = hash("" + keys[j]) % size;
                interim.push.apply(interim, array[i] || []);
            }

            /* Filter unique out */
            var result = [];
            interim.sort();
            length = interim.length;
            for (j = 0; j < length; j++) {
                if (interim[j - 1] !== interim[j])
                    result.push(interim[j]);
            }
            return result;
        };
    }

    /*
     * KubernetesWatch:
     * @api: a cockpit.http() object for the api server
     * @type: a string like 'pods' or 'nodes'
     * @update: invoked when ADDED or MODIFIED happens
     * @remove: invoked when DELETED happens
     *
     * Generates callbacks based on a Kubernetes watch.
     *
     * Each KubernetesWatch object watches a single type of object
     * in Kubernetes. The URI watched is /api/v1beta3/watch/<type>
     *
     * In addition to the above noted invocations of the callbacks,
     * if there is an ERROR, we restart the watch and invoke the
     * @remove callback with a null argument to indicate we are
     * starting over.
     */
    function KubernetesWatch(api, type, update, remove) {
        var self = this;

        var lastResource;
        var stopping = false;
        var requested = false;
        var req = null;

        /*
         * Each change is sent as an individual line from Kubernetes
         * but they may not arrive exactly that way, so we buffer
         * and split lines again.
         */

        var buffer;
        function handle_watch(data) {
            if (buffer)
                data = buffer + data;

            var lines = data.split("\n");
            var i, length = lines.length - 1;

            /* Last line is incomplete save for later */
            buffer = lines[length];

            /* Process all the others */
            var action, object;
            for (i = 0; i < length; i++) {
                try {
                    action = JSON.parse(lines[i]);
                } catch (ex) {
                    failure(ex);
                    req.close();
                    continue;
                }

                object = action.object;
                if (!object) {
                    console.warn("invalid watch without object");
                    continue;
                }

                if (object.metadata)
                    lastResource = object.metadata.resourceVersion;
                else
                    lastResource = null;

                if (action.type == "ADDED") {
                    update(object);
                } else if (action.type == "MODIFIED") {
                    update(object);
                } else if (action.type == "DELETED") {
                    remove(object);

                /* The watch failed, likely due to invalid resourceVersion */
                } else if (action.type == "ERROR") {
                    if (lastResource) {
                        lastResource = null;
                        start_watch();
                    }

                } else {
                    console.warn("invalid watch action type: " + action.type);
                    continue;
                }
            }
        }

        function start_watch() {
            var uri = "/api/v1beta3/watch/" + type;

            /*
             * If we have a last resource we can guarantee that we don't miss
             * any objects or changes to objects. If we don't have one, then we
             * have to list everything again. Still watch at the same time though.
             */
            if (requested && lastResource)
                uri += "?resourceVersion=" + encodeURIComponent(lastResource);

            /* Tell caller to remove all sources */
            else if (!requested)
                remove(null);

            /*
             * As a precaution, watch must take at least 1 second
             * to complete. Otherwise we could be in a tight loop here.
             * eg: if the API of Kubernetes changes unpredictably.
             */
            var waited = false;
            window.setTimeout(function() {
                waited = true;
            }, 1000);

            if (req) {
                req.cancelled = true;
                req.close();
            }

            req = api.get(uri)
                .stream(handle_watch)
                .fail(function(ex) {
                    req = null;
                    if (!stopping)
                        console.warn("watching kubernetes " + type + " failed: " + ex);
                })
                .done(function(data) {
                    var cancelled = req && req.cancelled;
                    req = null;
                    if (!stopping && !cancelled) {
                        if (!waited)
                            console.warn("watching kubernetes " + type + " didn't block");
                        else
                            start_watch();
                    }
                });
            requested = true;
        }

        start_watch();

        self.stop = function stop() {
            stopping = true;
            if (req)
                req.close();
        };
    }

    /**
     * KubernetesClient
     *
     * Properties:
     *  * objects: a dict of all the loaded kubernetes objects,
     *             with the 'uid' as the key
     *  * resourceVersion: latest resourceVersion seen
     */
    function KubernetesClient() {
        var self = this;

        var api = cockpit.http(8080);
        var watches = [ ];
        self.objects = { };

        /* TODO: Derive this value from cluster size */
        var index = new HashIndex(262139);

        self.resourceVersion = null;

        /*
         * Here we create KubernetesWatch object, and a property
         * definition for each type of object and hook them together.
         */
        function bind(type) {
            var flat = null;
            var items = { };
            var timeout;

            /* Always delay the update event a bit */
            function trigger() {
                flat = null;
                if (!timeout) {
                    timeout = window.setTimeout(function() {
                        timeout = null;
                        $(self).triggerHandler(type);
                    }, 100);
                }
            }

            function updated(item) {
                var meta = item.metadata;
                if (!meta || !meta.uid) {
                    console.warn("kubernetes item without uid");
                    return;
                }
                if (meta.resourceVersion && meta.resourceVersion > self.resourceVersion)
                    self.resourceVersion = meta.resourceVersion;

                items[meta.uid] = item;
                self.objects[meta.uid] = item;

                /* Add various bits to index, for quick lookup */
                var i, keys, length;
                if (meta.labels) {
                    keys = [];
                    for (i in meta.labels)
                        keys.push(i + meta.labels[i]);
                    index.add(keys, meta.uid);
                }
                var spec = item.spec;
                if (spec && spec.selector) {
                    keys = [];
                    for (i in spec.selector)
                        keys.push(i + spec.selector[i]);
                    index.add(keys, meta.uid);
                }

                /* Index the host for quick lookup */
                var status = item.status;
                if (status && status.host)
                    index.add([ status.host ], meta.uid);

                trigger();
            }

            function removed(item) {
                var key;
                if (!item) {
                    for (key in items) {
                        delete items[key];
                        delete self.objects[key];
                    }
                } else {
                    key = item.metadata ? item.metadata.uid : null;
                    if (!key) {
                        console.warn("kubernetes item without uid");
                        return;
                    }
                    delete items[key];
                    delete self.objects[key];
                }
                trigger();
            }

            Object.defineProperty(self, type, {
                enumerable: true,
                get: function get() {
                    if (!flat) {
                        flat = [];
                        for (var key in items)
                            flat.push(items[key]);
                        flat.sort(function(a1, a2) {
                            return (a1.metadata.name || "").localeCompare(a2.metadata.name || "");
                        });
                    }
                    return flat;
                }
            });

            var wc = new KubernetesWatch(api, type, updated, removed);
            watches.push(wc);
        }

        /*
         * Define and bind various properties which are arrays of objects
         *
         * TODO: Decide whether we need these as exposed properties, or
         * whether callers can just use the .select() method.
         */

        bind("nodes");
        bind("pods");
        bind("services");
        bind("replicationcontrollers");

        /**
         * client.select()
         * @selector: plain javascript object, JSON label selector
         * @type: optional kind string (eg: 'Pod')
         *
         * Select objects that match the given labels.
         *
         * Returns: an array of objects
         */
        this.select = function select(selector, kind) {
            var i, keys;
            var possible, match, results = [];
            if (selector) {
                keys = [];
                for (i in selector)
                    keys.push(i + selector[i]);
                possible = index.select(keys);
            } else {
                possible = Object.keys(self.objects);
            }
            var obj, uid, j, length = possible.length;
            for (j = 0; j < length; j++) {
                uid = possible[j];
                obj = self.objects[uid];
                if (!obj || !obj.metadata)
                    continue;
                if (selector && !obj.metadata.labels)
                    continue;
                if (kind && obj.kind !== kind)
                    continue;
                match = true;
                for (i in selector) {
                    if (obj.metadata.labels[i] !== selector[i]) {
                        match = false;
                        break;
                    }
                }
                if (match)
                    results.push(obj);
            }
            return results;
        };

        /**
         * client.infer()
         * @labels: plain javascript object, JSON labels
         * @kind: optional kind string
         *
         * Infer which objects that have selectors would have
         * matched the given labels.
         */
        this.infer = function infer(labels, kind) {
            var i, keys;
            var possible, match, results = [];
            if (labels) {
                keys = [];
                for (i in labels)
                    keys.push(i + labels[i]);
                possible = index.select(keys);
            } else {
                possible = Object.keys(self.objects);
            }
            var obj, uid, j, length = possible.length;
            for (j = 0; j < length; j++) {
                uid = possible[j];
                obj = self.objects[uid];
                if (!obj || !obj.spec || !obj.spec.selector)
                    continue;
                if (kind && obj.kind !== kind)
                    continue;
                match = true;
                if (labels) {
                    for (i in obj.spec.selector) {
                        if (labels[i] !== obj.spec.selector[i]) {
                            match = false;
                            break;
                        }
                    }
                }
                if (match)
                    results.push(obj);
            }
            return results;
        };

        /**
         * client.hosting()
         * @host: the node host name, required
         *
         * Find out which objects are being hosted at the given node. These
         * have a obj.status.host property equal to the @host name passed into
         * this function.
         *
         * Returns: an array of kubernetes objects
         */
        this.hosting = function hosting(host) {
            var possible = index.select([ host ]);
            var obj, j, length = possible.length;
            var results = [];
            for (j = 0; j < length; j++) {
                obj = self.objects[possible[j]];
                if (obj && obj.status && obj.status.host === host)
                    results.push(obj);
            }
            return results;
        };

        this.delete_pod = function delete_pod(pod_name) {
            api.request({"method": "DELETE",
                "body": "",
                "path": "/api/v1beta3/namespaces/default/pods/" + encodeURIComponent(pod_name)
            }).fail(failure);
        };

        this.delete_node = function delete_node(nodes_name) {
            api.request({"method": "DELETE",
                "body": "",
                "path": "/api/v1beta3/namespaces/default/nodes/" + encodeURIComponent(nodes_name)
            }).fail(failure);
        };

        this.delete_replicationcontroller = function delete_replicationcontroller(rc_name) {
            api.request({"method": "DELETE",
                "body": "",
                "path": "/api/v1beta3/namespaces/default/replicationcontrollers/" + encodeURIComponent(rc_name)
            }).fail(failure);
        };

        this.delete_service = function delete_service(service_name) {
            api.request({"method": "DELETE",
                "body": "",
                "path": "/api/v1beta3/namespaces/default/services/" + encodeURIComponent(service_name)
            }).fail(failure);
        };

        this.create_replicationcontroller = function create_replicationcontroller(replicationcontroller_json) {
            api.post("/api/v1beta3/namespaces/default/replicationcontrollers", replicationcontroller_json)
               .fail(failure);
        };

        this.create_node = function create_node(node_json) {
            api.post("/api/v1beta3/namespaces/default/nodes", node_json)
               .fail(failure);
        };

        this.create_pod = function create_pod(pod_json) {
            api.post("/api/v1beta3/namespaces/default/pods", pod_json)
               .fail(failure);
        };

        this.create_service = function create_pod(service_json) {
            api.post("/api/v1beta3/namespaces/default/services", service_json)
               .fail(failure);
        };

        this.update_replicationcontroller = function update_replicationcontroller(rc_json, rc_name) {
            api.request({"method": "PUT",
                "body": rc_json,
                "path": "/api/v1beta3/namespaces/default/replicationcontrollers/" + encodeURIComponent(rc_name)
            }).fail(failure);
        };

        self.close = function close() {
            var w = watches;
            watches = [ ];
            $.each(w, function(i, wc) {
                wc.stop();
            });
        };
    }

    function EtcdClient() {
        var self = this;

        var etcd_api = cockpit.http(7001);
        var first = true;
        var later;

        function receive(data, what ,kind) {
            var resp = JSON.parse(data);
            self[what] = resp;

            if (!first)
                $(self).triggerHandler(what, [ self[what] ]);
        }

        function update() {
            var reqs = [];

            reqs.push(etcd_api.get("/v2/admin/machines")
                .fail(failure)
                .done(function(data) {
                    receive(data, "etcdHosts");
                }));

            reqs.push(etcd_api.get("/v2/keys/coreos.com/network/config")
                .fail(failure)
                .done(function(data) {
                    receive(data, "flannelConfig");
                }));

            if (first) {
                $.when.apply($, reqs)
                    .always(function() {
                        first = false;
                        $(self).triggerHandler("etcdHosts", [ self.etcdHosts ]);
                    });
            }
        }

        update();
    }

    kubernetes.k8client = function client() {
        return new KubernetesClient();
    };

    kubernetes.etcdclient = function client() {
        return new EtcdClient();
    };

    return kubernetes;
});
